# Copyright 2009-2013, Simon Kennedy, code@sffjunkie.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a dictionary of validation functions to test values. Values
are checked that they...

:str: can be converted to a string 
:bool: can be converted to a boolean value
:int: can be converted to an integer value
:float: can be converted to a floating point value
:yesno: match one of 'yes', 'y', 'no', 'n' with any case plus True and False
:re(regexp): matches the regular expression `regexp`
:path: is a valid path name that exists
:path(empty): is a valid path name that is empty
:path(nonempty): is a valid path name that is not empty
:path(pathspec): is a valid path name that contains files that conform to `filespec`

`pathspec` is of the form [+-]glob where the leading '+' indicates that
the path must include a file that matches the pattern and '-' indicates that
it must not include files that match pattern. Multiple pathspecs can be
specified separated by commas.
"""

import re
import csv
import glob
from io import StringIO
from os import path, listdir

try:
    from pkg_resources import load_entry_point
except:
    load_entry_point = lambda x: None

__all__ = ['Validators']

def validate_str(value):
    return str(value)

def validate_bool(value):
    true_values = ['true', '1', 'yes', 'y']
    false_values = ['false', '0', 'no', 'n']
    
    if isinstance(value, bool):
        return value
    elif isinstance(value, int):
        return bool(value)
    elif str(value).lower() in true_values:
        return True
    elif str(value).lower() in false_values:
        return False
    else:
        raise ValueError('Please enter a valid boolean value')
    
def validate_int(value):
    if value == '':
        return 0
    
    try:
        return int(value)
    except:
        raise ValueError('Please enter an integer value')
    
def validate_float(value):
    if value == '':
        return 0.0
    
    try:
        return float(value)
    except:
        raise ValueError('Please enter a floating point value')

def validate_regex(regex):
    def validate(value):
        r = re.compile(regex)
        m = r.match(value)
        if m is not None:
            return value
        else:
            raise ValueError(('Please enter a string which matches'
                                   'the regular expression %s' % regex))
    
    return validate

def validate_path(value):
    is_dir = path.exists(value) and path.isdir(value)
    if not is_dir:
        raise ValueError('Please enter a valid path name.')
    return value

def validate_path_empty(value):
    is_dir = path.exists(value) and path.isdir(value)
    if not is_dir:
        raise ValueError(('Please enter a valid path name.'))
    
    if len(listdir(value)) != 0:
        raise ValueError(('Please enter a valid path name '
                               'for which the path is empty.'))
    return value

def validate_path_nonempty(value):
    is_dir = path.exists(value) and path.isdir(value)
    if not is_dir:
        raise ValueError(('Please enter a valid path name.'))
    
    if len(listdir(value)) == 0:
        raise ValueError(('Please enter a valid path name '
                          'for which the path is not empty.'))
    return value

def path_includes_file(path, filespec):
    fname = path.join(path, filespec)
    return path.exists(fname)

def path_does_not_include_file(path, filespec):
    fname = path.join(path, filespec)
    return not path.exists(fname)

def validate_path_with_spec(pathspec):
    reader = csv.reader(StringIO(unicode(pathspec)))
    included = []
    not_included = []
    for row in reader:
        for elem in row:
            if elem.startswith('+'):
                included.append(elem[1:].strip('"'))
            elif elem.startswith('-'):
                not_included.append(elem[1:].strip('"'))

    def validate(value):
        not_found = []
        for spec in included:
            if len(glob.glob(path.join(value, spec))) == 0:
                not_found.append(spec)
        
        found = []
        for spec in not_included:
            if len(glob.glob(path.join(value, spec))) != 0:
                found.append(spec)
        
        found_count = len(found) 
        not_found_count = len(not_found) 
        if found_count != 0 or not_found_count != 0:
            msg_elem = ['Directory %s' % value]
            if not_found_count > 0:
                msg_elem.append('should contain files matching %s' % ','.join(included))
                
            if found_count > 0:
                if not_found_count > 0:
                    msg_elem.append('and')

                msg_elem.append('should not contain files matching %s' % ','.join(not_included))
                
            msg = ' '.join(msg_elem)
            raise ValueError(msg)
            
        return value
    
    return validate

def validate_nonempty(value):
    if value is None or str(value) == '':
        raise ValueError("Please enter something in this field.")
    return value

class _Registry():
    def __init__(self):
        self._validators = {
            'str': lambda value: str(value),
            'yesno': validate_bool,
            'bool': validate_bool,
            'int': validate_int,
            'float': validate_float,
            # 're(regex)'
            'path': validate_path,
            'path(empty)': validate_path_empty,
            'path(nonempty)': validate_path_nonempty,
            # 'path(pathspec)'
            'nonempty': validate_nonempty,
        }
        
        self._entry_point_re = re.compile('\w+(\.\w)?\:\w+(\.\w)?')

    def __getitem__(self, key):
        if key is None:
            return None
        
        if key in self._validators:
            return self._validators[key]
        
        if key.startswith('re(') and key.endswith(')'):
            return validate_regex(key[3:-1])
        
        if key.startswith('path(') and key.endswith(')'):
            filespec = key[5:-1]
            return validate_path_with_spec(filespec)
        
        ep_match = self._entry_point_re.match(key)
        if ep_match is not None:
            func = load_entry_point(key)
            self._validators[key] = func
            return func
        
        return None

Validators = _Registry()



# Copyright 2009-2013, Simon Kennedy, code@sffjunkie.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import readline
except ImportError:
    pass

try:
    # Try assigning the Python 2 version of the input function
    input = raw_input
except:
    pass

from os.path import normpath
from collections import OrderedDict

from ama import Asker
from ama.validator import validate_bool
from ama.color import bright_green, cyan, bright_red

class TerminalAsker(Asker):
    """Ask the questions by using the terminal"""
    
    def __init__(self, title, preamble='', filename=''):
        Asker.__init__(self, filename)
        self._title = title
        self._preamble = preamble
        self._ask = OrderedDict()
    
    def add_question(self, key, question):
        """Add a question to the list of questions.
        
        Called by the :meth:`Asker.ask` method or by your code.
        """
        
        self._ask[key] = question
    
    def go(self, initial_answers):
        """Perform the question asking"""
        
        print(bright_green(self._title))
        print(bright_green('-'*len(self._title)))
        if self._preamble != '':
            print('\n%s' % self._preamble)
        print('\n')
           
        result = {}
        try:
            answers = {}
            for key, question in self._ask.items():
                answer = self._ask_question(question, initial_answers, answers)
                answers[key] = answer
                
            result[u'answers'] = answers
            result[u'valid'] = True
            result[u'result'] = 'ok'
        except (KeyboardInterrupt, EOFError):
            try:
                result[u'result'] = 'cancel'
                print('\n[Interrupted]\n')
            except:
                pass
            
        return result

    def _ask_question(self, question, initial_answers, answers):
        """Ask a single question"""
        
        prompt_tail = ''
        if question.validator == 'yesno':
            if not question.label.endswith(' (y/n)'):
                prompt_tail = ' (y/n)'
            if question.default == True or question.default == 'y':
                prompt_tail += ' [y]'
            elif question.default == False or question.default == 'n':
                prompt_tail += ' [n]'
            default = validate_bool(question.default)
        else:
            try:
                default = question.default.format(**answers)
            except:
                default = question.default
            
            if question.validator and question.validator.startswith('path'):
                default = normpath(default)
            
            prompt_tail = ' [%s]' % default
        
        prompt = '%s%s: ' % (question.label, prompt_tail)
            
        while True:
            answer = input(cyan(prompt))
            stripped = answer.strip()
            if stripped == '?':
                print(question.help_text)
                continue
            
            if stripped == '':
                answer = default
                
            try:
                validate = self.validator(question.type, question.validator)
                answer = validate(answer)
            except Exception as err:
                print(bright_red('* %s\n' % str(err)))
                print(question.help_text)
                continue
            break
    
        return answer



# Copyright 2009-2013, Simon Kennedy, code@sffjunkie.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from string import Formatter
from collections import namedtuple, OrderedDict

from ama.validator import Validators

import sys
if sys.version < '3':
    import codecs
    def u(x):
        return codecs.unicode_escape_decode(x)[0]
else:
    def u(x):
        return x
        
_Question = namedtuple('question',
                       'key label type default help_text validator')

def Question(key, label, answer_type='str', default=None, help_text='',
             validator=None):
    return _Question(key, label, answer_type, default, help_text, validator)


class Asker(object):
    def __init__(self, filename=''):
        if filename != '':
            with open(filename) as fp:
                data = fp.read()
                self._questions = json.loads(data,
                                             object_pairs_hook=OrderedDict)
        else:
            self._questions = None
            
        self.depends_on = {}
        self.depends_on_us = {}
    
    def ask(self, questions=None, initial_answers=None, all_questions=True,
            validators={}):
        """Ask the questions and return the answers
        
        :param questions: The questions to prompt for answers. Can either be
                          a json formatted string or a dict subclass
        :type questions:  string or dict
        :param initial_answers:   A dictionary containing the already answered
                          questions
        :type initial_answers:    dict
        :param all_questions: If True only the already unanswered questions
                              will be asked; if False all questions will be
                              asked.
        :type all_questions:  bool
        :param validators: A dictionary of custom validator functions
        :type validators: dict
        :returns: The answers
        :rtype: OrderedDict
        """
    
        if questions is None and self._questions is not None:
            questions = self._questions
            
        if not isinstance(questions, dict):
            questions = json.loads(questions, object_pairs_hook=OrderedDict)
    
        if initial_answers is None:
            initial_answers = {}
        
        for key, question in questions.items():
            if all_questions or key not in initial_answers:
                default = initial_answers.get(key, question[2])
                
                depends_on = self._find_dependencies(default)
                if len(depends_on) != 0:
                    self.depends_on[key] = depends_on
                    for dep in depends_on:
                        if dep not in self.depends_on_us:
                            self.depends_on_us[dep] = []
                        self.depends_on_us[dep].append(key)
                
                q = Question(key, question[0], question[1],
                             default, question[3], question[4])
                self.add_question(key, q)

        result = self.go(initial_answers)
        return result

    def go(self):
        """Overridden by subclasses to ask the questions. Subclasses should
        return a dictionary of the answers"""
        
        raise NotImplemented

    def validator(self, type_validator, custom_validator=None):
        """Returns a function which validates by type first and then by a custom
        validator if one is provided
        """
        
        def validate(value):
            tv = Validators[type_validator]
            
            cv = None
            if custom_validator is not None:
                cv = Validators[custom_validator]
            
            if tv is None and cv is None:
                return value
            else:
                v1 = tv(value)
                if cv is not None:
                    return cv(v1)
                else:
                    return v1

        return validate
    
    def _find_dependencies(self, default):
        """Returns a list of question keys the default value uses""" 
        
        f = Formatter()
        dependencies = []
        for _text, field, _format_spec, _conversion in f.parse(str(default)):
            if field is not None:
                dependencies.append(field)
        
        return dependencies
    


## {{{ http://code.activestate.com/recipes/576688/ (r1)
try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

from time import time

class ToolTip(tk.Toplevel):
    """
    Provides a ToolTip widget for Tkinter.
    To apply a ToolTip to any Tkinter widget, simply pass the widget to the
    ToolTip constructor
    """ 
    def __init__( self, wdgt, msg=None, msgFunc=None, delay=1, follow=True ):
        """
        Initialize the ToolTip
        
        Arguments:
          wdgt: The widget this ToolTip is assigned to
          msg:  A static string message assigned to the ToolTip
          msgFunc: A function that retrieves a string to use as the ToolTip text
          delay:   The delay in seconds before the ToolTip appears(may be float)
          follow:  If True, the ToolTip follows motion, otherwise hides
        """
        
        self.wdgt = wdgt
        # The parent of the ToolTip is the parent of the ToolTips widget
        self.parent = self.wdgt.master
        
        # Initalise the Toplevel                                          
        tk.Toplevel.__init__(self, self.parent, bg='black', padx=1, pady=1)
        self.withdraw()                                                         # Hide initially
        self.overrideredirect( True )                                           # The ToolTip Toplevel should have no frame or title bar
        
        self.msgVar = tk.StringVar()                                               # The msgVar will contain the text displayed by the ToolTip        
        if msg == None:                                                         
            self.msgVar.set('No message provided')
        else:
            self.msgVar.set(msg)
            
        self.msgFunc = msgFunc
        self.delay = delay
        self.follow = follow
        self.visible = 0
        self.lastMotion = 0
        # The test of the ToolTip is displayed in a Message widget
        tk.Message( self, textvariable=self.msgVar, bg='#FFFFDD',
                 aspect=1000 ).grid()
        
        # Add bindings to the widget.  This will NOT override bindings that the widget already has
        self.wdgt.bind( '<Enter>', self.spawn, '+' )
        self.wdgt.bind( '<Leave>', self.hide, '+' )
        self.wdgt.bind( '<Motion>', self.move, '+' )
        
    def spawn( self, event=None ):
        """
        Spawn the ToolTip.  This simply makes the ToolTip eligible for display.
        Usually this is caused by entering the widget
        
        Arguments:
          event: The event that called this funciton
        """
        self.visible = 1
        # The after function takes a time argument in miliseconds
        self.after(int(self.delay * 1000), self.show)
        
    def show( self ):
        """
        Displays the ToolTip if the time delay has been long enough
        """
        if self.visible == 1 and time() - self.lastMotion > self.delay:
            self.visible = 2
        if self.visible == 2:
            self.deiconify()
            
    def move( self, event ):
        """
        Processes motion within the widget.
        
        Arguments:
          event: The event that called this function
        """
        self.lastMotion = time()
        # If the follow flag is not set, motion within the widget will make the ToolTip dissapear
        if self.follow == False:
            self.withdraw()
            self.visible = 1
        # Offset the ToolTip 10x10 pixes southwest of the pointer
        self.geometry( '+%i+%i' % ( event.x_root+10, event.y_root+10 ) )
        try:
            # Try to call the message function.  Will not change the message
            # if the message function is None or the message function fails
            self.msgVar.set( self.msgFunc() )
        except:
            pass
        self.after( int( self.delay * 1000 ), self.show )
            
    def hide( self, event=None ):
        """
        Hides the ToolTip.  Usually this is caused by leaving the widget
        
        Arguments:
          event: The event that called this function
        """
        self.visible = 0
        self.withdraw()
## end of http://code.activestate.com/recipes/576688/ }}}



# Copyright 2009-2013, Simon Kennedy, code@sffjunkie.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os.path
from collections import OrderedDict

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

try:
    from tkinter import font
except ImportError:
    import tkFont as font

try:
    from tkinter import ttk
except ImportError:
    import ttk
    
try:
    from tkinter import filedialog
except ImportError:
    import tkFileDialog as filedialog

from ama import Asker, u
from ama.tk_tooltip import ToolTip

class TkAsker(Asker):
    def __init__(self, title, preamble='', filename='', allow_invalid=True):
        Asker.__init__(self, filename)
        self._title = title
        self._preamble = preamble
        self._allow_invalid = allow_invalid
        self._row = 0
        self._ask = OrderedDict()

        self._root = tk.Tk()
        self._root.title(self._title)
        
        self._edited_entry = ttk.Style()
        self._edited_entry.configure('edited.TEntry', foreground='black')
        
        self._unedited_entry = ttk.Style()
        self._unedited_entry.configure('unedited.TEntry', foreground='#666')

        header_font = font.Font(family='TkDefaultFont')
        header_font.configure(weight='bold')
        header = ttk.Label(self._root, text=self._preamble, padding=3,
                           font=header_font, wraplength=420)
        header.grid(column=0, row=0, sticky=(tk.N, tk.EW))
        
        self.content = ttk.Frame(self._root, padding=(3,3,3,3))
        self.content.grid(column=0, row=1, sticky=(tk.N, tk.S, tk.E, tk.W))
        self.content.columnconfigure(0, weight=1)

        self.content.columnconfigure(0, weight=0)
        self.content.columnconfigure(1, weight=1)
        self.content.columnconfigure(2, weight=0)
        self.content.columnconfigure(3, weight=0)
        
        self._root.columnconfigure(0, weight=1)
        self._root.rowconfigure(0, weight=0)
        
        okcancel = ttk.Frame(self._root, padding=(3,3,3,3))
        
        # Swap the order of buttons for Windows
        if sys.platform.startswith('win32'):
            btn_column = (1,2)
        else:
            btn_column = (2,1)
            
        self.ok_btn = ttk.Button(okcancel, text='OK', width=10, command=self._ok)
        self.ok_btn.grid(column=btn_column[0], row=0, padx=(6, 0))
        cancel = ttk.Button(okcancel, text='Cancel', width=10,
                            command=self._cancel)
        cancel.grid(column=btn_column[1], row=0, padx=(6, 0))

        okcancel.columnconfigure(0, weight=1)
        okcancel.columnconfigure(1, weight=0)
        okcancel.columnconfigure(2, weight=0)
        
        okcancel.grid(column=0, row=2, sticky=(tk.E, tk.W, tk.S))
        self._root.rowconfigure(2, weight=1)
        
        self._root.protocol('WM_DELETE_WINDOW', self._cancel)
        
        if sys.platform.startswith('darwin'):
            self._root.createcommand("::tk::mac::Quit", self._cancel)
    
    def add_question(self, key, question):
        tkq = TkQuestion(self, self._row,
                         question)
        self._ask[key] = tkq
        self._row = self._row + 1

    def current_answers(self, update_info=None):
        current_answers = {}
        for key, tkq in self._ask.items():
            current_answers[key] = tkq.value
        
        if update_info is not None:
            current_answers[update_info[0]] = update_info[1]
            
        return current_answers
    
    def update_answers(self, update_info=None):
        answers = self.current_answers(update_info)
            
        for key, tkq in self._ask.items():
            if (update_info is None or key != update_info[0]) and \
            (not tkq.edited or key == update_info[0]):
                tkq.update(answers)

    def go(self, initial_answers):
        self._result = {}
        self.update_answers()
        self._root.mainloop()
        return self._result

    def check_invalid(self):
        if not self._allow_invalid and not self._is_valid():
            self.ok_btn.state(['disabled'])
        else:
            self.ok_btn.state(['!disabled'])
            
        self.ok_btn.update_idletasks()

    def _is_valid(self):
        for _key, question in self._ask.items():
            if not question.valid:
                return False
        
        return True
    
    def _ok(self, event=None):
        self._result['valid'] = self._is_valid()
        self._result['result'] = 'ok'
        
        answers = {}
        for key, tkq in self._ask.items():
            answers[key] = tkq.value
            
        self._result['answers'] = answers
        
        self._root.destroy()
    
    def _cancel(self, event=None):
        self._result['valid'] = False
        self._result['result'] = 'cancel'
        self._root.destroy()


class TkQuestion(object):
    def __init__(self, asker, row, question):
        self._key = question.key
        self._label = question.label
        self._type = question.type
        self._default = question.default
        self._validator = question.validator
        if self._validator and self._validator.startswith('path'):
            self._default = os.path.normpath(self._default)
            
        self._validate = asker.validator(self._type, self._validator)

        self._var = None
        self._asker = asker
        self._row = row
        self._entry = None
        
        self._is_edited = False
        self._is_valid = True

        self.label = ttk.Label(asker.content, text=self._label)
        self.label.grid(column=0, row=self._row, sticky=(tk.N, tk.S, tk.W),
                        padx=(0,5))
        
        error_font = font.Font(family='TkFixedFont', size=10, weight='bold')
        self.info_label = ttk.Label(asker.content, font=error_font, width=2,
                                    anchor=tk.CENTER)
        self.info_label.grid(column=2, row=self._row, padx=(3,0))
        self._help_text = question.help_text
        if self._help_text != '':
            self.info_label['text'] = '?'
            self.tooltip = ToolTip(self.info_label, msg=self._help_text, delay=0.5)
        
        self._validate_entry = (asker._root.register(self._tk_validate),
                                '%P', '%V')

        current_answers = self._asker.current_answers()

        if self._validator and self._validator.startswith('path'):
            self._var = tk.StringVar()
            frame = ttk.Frame(asker.content)
            self._entry = ttk.Entry(frame, textvariable=self._var,
                                   validate='all',
                                   validatecommand=self._validate_entry)
            self._entry.grid(column=0, row=0, sticky=tk.EW)
            btn = ttk.Button(frame, text='Browse...',
                             command=self._browse_for_directory)
            btn.grid(column=1, row=0, sticky=tk.E)
            frame.columnconfigure(0, weight=1)
            frame.columnconfigure(1, weight=0)

            self.update(current_answers)
            
        elif self._type == 'str':
            self._var = tk.StringVar()
            self._entry = ttk.Entry(asker.content, textvariable=self._var,
                                   validate='all',
                                   validatecommand=self._validate_entry)
            frame = self._entry
            self.update(current_answers)
            
        elif self._type == 'int' or isinstance(self._type, int):
            self._var = tk.IntVar()
            self._entry = ttk.Entry(asker.content, textvariable=self._var,
                                   validate='all',
                                   validatecommand=self._validate_entry)
            self._entry.configure(width=30)
            frame = self._entry

            self.update(current_answers)
            
        elif self._type == 'float' or isinstance(self._type, float):
            self._var = tk.DoubleVar()
            self._entry = ttk.Entry(asker.content, textvariable=self._var,
                                   validate='all',
                                   validatecommand=self._validate_entry)
            self._entry.configure(width=30)
            frame = self._entry

            self.update(current_answers)
            
        elif self._type == 'bool' or self._validator == 'yesno' or \
            isinstance(self._type, bool):
            self._var = tk.BooleanVar()
            self.value = self._default
            frame = ttk.Frame(asker.content)
            
            if self._validator == 'yesno':
                text = ('Yes', 'No')
            else:
                text = ('True', 'False')
                
            y = ttk.Radiobutton(frame, text=text[0],
                                variable=self._var, value=True)
            y.grid(column=0, row=0, padx=(0,5))
            
            n = ttk.Radiobutton(frame, text=text[1],
                                variable=self._var, value=False)
            n.grid(column=1, row=0)

        elif isinstance(self._type, list):
            self._var = tk.StringVar()
            self.value = self._type[0]
            if len(self._type) <= 3:
                frame = ttk.Frame(asker.content)
                for idx, e in enumerate(self._type):
                    rb = ttk.Radiobutton(frame, text=str(e),
                                         variable=self._var, value=str(e))
                    rb.grid(column=idx, row=0, padx=(0,5))
            else:
                self._entry = ttk.Combobox(asker.content,
                                          textvariable=self._var)
                self._entry['values'] = tuple(self._type)
                frame = self._entry
            
        else:
            raise ValueError(('Unable to create _entry widget '
                              'for type %s') % self._type)
            
        frame.grid(column=1, row=self._row, sticky=tk.EW, padx=0)
        
        self.edited = self._is_edited
        asker.content.rowconfigure(self._row, weight=0)
        
    def update(self, current_answers):
        updated_answer = str(self._default).format(**current_answers)
        self.value = updated_answer
        
    def value():
        def fget(self):
            try:
                return self._validate(self._var.get())
            except:
                return self._entry.get()
        
        def fset(self, value):
            try:
                value = self._validate(value)
                self._var.set(value)
                self.valid = True
            except ValueError:
                self._var.set(value)
                self.valid = False
                
        return locals()
    
    value = property(**value())

    def valid():
        def fget(self):
            return self._is_valid
        
        def fset(self, value):
            self._is_valid = bool(value)
            if self._is_valid:
                if self._help_text != '':
                    self.info_label['text'] = '?'
                self.info_label['foreground'] = 'black'
            else:
                self.info_label['text'] = '!'
                self.info_label['foreground'] = 'red'
                
        return locals()

    valid = property(**valid())
        
    def edited():
        def fget(self):
            return self._is_edited
        
        def fset(self, value):
            self._is_edited = bool(value)
            
            if isinstance(self._entry, ttk.Entry):
                if self._is_edited:
                    self._entry['style'] = 'edited.TEntry'
                else:
                    self._entry['style'] = 'unedited.TEntry'
                    
        return locals()
    
    edited = property(**edited())
    
    def _tk_validate(self, P, V):
        rtn = 1
        if V == 'focusout':
            if P.strip() == '':
                self.valid = True
                self.edited = False
            else:
                try:
                    _value = self._validate(P)
                    self.valid = True
                except:
                    self.valid = False
                    rtn = 0
                
            self._asker.check_invalid()
        elif V == 'key':
            if not self.edited:
                self.edited = True
            
            self._asker.update_answers((self._key, P))

        return rtn
    
    def _browse_for_directory(self, *args):
        path_entry = self.value
        path_entry = os.path.abspath(os.path.expanduser(path_entry))
        
        new_path = filedialog.askdirectory(initialdir=path_entry)
        new_path = os.path.normpath(new_path)
        
        home_path = u(os.path.expanduser('~'))
        current_path = u(os.getcwd())
        
        if new_path == home_path:
            new_path = '~'
        elif new_path.startswith(home_path):
            new_path = os.path.join('~', new_path[len(home_path)+1:])
        elif new_path == current_path:
            new_path = '.'
        elif new_path.startswith(current_path):
            new_path = os.path.join('.', new_path[len(current_path)+1:])
            
        self.value = new_path
        self._asker.update_answers((self._key, new_path))



# Copyright 2009-2013, Simon Kennedy, code@sffjunkie.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import colorama
    colorama.init(autoreset=True)
    COLOR = True
except:
    COLOR = False
    
def colorize(intensity, color, text):
    if COLOR:
        return colorama.Style.__dict__[intensity.upper()] + colorama.Fore.__dict__[color.upper()] + text
    else:
        return text

def create_color_func(intensity, name):
    def inner(text):
        return colorize(intensity, name, text)
    if intensity == 'normal':
        globals()[name] = inner
    else:
        globals()['%s_%s' % (intensity, name)] = inner

if COLOR:
    intensities = [x for x in colorama.Style.__dict__.keys() if x!='RESET_ALL']
    colours = [x for x in colorama.Fore.__dict__.keys() if x!='RESET']
else:
    intensities = ['DIM', 'NORMAL', 'BRIGHT']
    colours = ['BLACK', 'RED', 'GREEN', 'YELLOW', 'BLUE', 'MAGENTA', 'CYAN', 'WHITE']

for intensity in intensities:
    for fore in colours:
        create_color_func(intensity.lower(), fore.lower())
